import attr
import boto3
import semver
import yaml
from aws_conduit import conduit_factory as factory
from aws_conduit import helper
from aws_conduit.aws import s3, service_catalog


@attr.s
class ConduitProduct(yaml.YAMLObject):
    yaml_loader = yaml.SafeLoader
    yaml_tag = u'!Product'
    name = attr.ib()
    owner = attr.ib()
    bucket = attr.ib()
    cfn_type = attr.ib()
    portfolio = attr.ib()
    description = attr.ib(default='No description set')
    template = attr.ib(default=None)
    template_prefix = attr.ib(default=None)
    product_id = attr.ib(default=None)
    version = attr.ib(default="0.0.0")
    provisioned = attr.ib(default=[])
    role = attr.ib(default=None)
    resources = attr.ib(default=[])

    def _add_initial_template(self):
        template = dict(
            AWSTemplateFormatVersion="2010-09-09",
            Resources=dict(
                FalseS3=dict(
                    Type="AWS::S3::Bucket",
                    Properties=dict(
                        AccessControl="private"
                    )
                )
            ))
        self.bucket.put_config(template, self.template_prefix)

    def create(self, support, tags):
        """
        Create a new product.
        """
        if not self.template:
            self.template = "{}/{}/{}/{}/{}.{}".format(self.bucket.get_url(), self.portfolio, self.name, self.version, self.name, self.cfn_type)
            self.template_prefix = "{}/{}/{}/{}.{}".format(self.portfolio, self.name, self.version, self.name, self.cfn_type)
        self._add_initial_template()
        description = 'NotSet'
        email = 'noone@home.com'
        url = 'http://notset.com'
        if 'description' in support:
            description = support['description']
        if 'email' in support:
            email = support['email']
        if 'url' in support:
            url = support['url']
        create_response = service_catalog.create_product(self, email, url, description, [], self.template)
        self.product_id = create_response['ProductViewDetail']['ProductViewSummary']['ProductId']

    def create_role(self, name):
        print(self.role)
        if self.role is not None:
            if self.role.name != name:
                self.role = factory.role(name)
                self.role.create()
        else:
            self.role = factory.role(name)
            self.role.create()

    def add_to_portfolio(self, portfolio_id):
        # if not self.product_id:
        self.set_product_id()
        service_catalog.associate(self.product_id, portfolio_id)

    def set_product_id(self):
        if self.product_id is None:
            summary = self.get_summary()
            self.product_id = summary['ProductId']

    def get_summary(self):
        response = service_catalog.search(self.name)
        if 'ProductViewDetails' in response:
            for item in response['ProductViewDetails']:
                if item['ProductViewSummary']:
                    summary = item['ProductViewSummary']
                    if (summary['Name'] == self.name and
                            summary['Owner'] == self.owner):
                        return summary
        return None

    def exists(self):
        summary = self.get_summary()
        return bool(summary is not None)

    def delete(self):
        self.set_product_id()
        portfolios = self.get_all_portfolios()
        for portfolio in portfolios:
            self.disassociate(portfolio)
        self.service_catalog.delete_product(
            Id=self.product_id
        )

    def update(self, support):
        """
        Updates this Product instance.
        """
        description = 'NotSet'
        email = 'noone@home.com'
        url = 'http://notset.com'
        if 'description' in support:
            description = support['description']
        if 'email' in support:
            email = support['email']
        if 'url' in support:
            url = support['url']
        service_catalog.update_product(self.product_id,
                                       self.name,
                                       self.owner,
                                       self.description,
                                       description,
                                       email,
                                       url)

    def disassociate(self, portfolio):
        service_catalog.disassociate(self.product_id, portfolio)

    def get_all_portfolios(self):
        self.set_product_id()
        return service_catalog.list_portfolios_for_product(self.product_id)

    def release(self, release_type, local_template, current_version):
        product_version = current_version
        if release_type == 'build':
            product_version = semver.bump_build(current_version)
        if release_type == 'major':
            product_version = semver.bump_major(current_version)
        if release_type == 'minor':
            product_version = semver.bump_minor(current_version)
        if release_type == 'patch':
            product_version = semver.bump_patch(current_version)

        self.release_new_build(local_template, product_version)

    def release_new_build(self, local_template, product_version):
        helper.put_resource(local_template, local_template, self.bucket, self.portfolio, self.name, product_version, environment=None)
        for resource in self.resources:
            helper.put_resource(resource, resource, self.bucket, self.portfolio, self.name, product_version, environment=None)
        template_url = "{}/{}/{}/{}/{}".format(self.bucket.get_url(), self.portfolio, self.name, product_version, local_template)
        print("Creating new version to template: {}".format(template_url))
        service_catalog.new_version(self.product_id, product_version, template_url)
        self.version = product_version
        print("Released new product version: {}".format(product_version))

    def add_resources(self, product_spec):
        self.resources = []
        if 'associatedResources' in product_spec:
            for resource in product_spec['associatedResources']:
                self.resources.append(resource)
        if 'nestedStacks' in product_spec:
            for resource in product_spec['nestedStacks']:
                self.resources.append(resource)

    def tidy_versions(self):
        versions = self.get_all_versions()
        version = self.version
        for item in versions:
            if ('build' in item['Name'] and
                    semver.compare(item['Name'], version) == -1):
                self.delete_version(item['Name'], item['Id'])
        print("Current product version is: {}".format(version))
        return version

    def delete_version(self, version_name, version_id):
        print("Deleting version: {}".format(version_name))
        service_catalog.delete_version(self.product_id, version_id)
        prefix = "{}/{}/{}".format(self.portfolio, self.name, version_name)
        s3.delete_folder(self.bucket.name, prefix)

    def get_all_versions(self):
        return service_catalog.list_all_versions(self.product_id)

    def get_last_version(self):
        versions = self.get_all_versions()
        version = self.version
        for item in versions:
            if semver.compare(item['Name'], version) == -1:
                version = item['Name']
                break
        print("Current product version is: {}".format(version))
        return version

    def get_version_id(self):
        versions = self.get_all_versions()
        for item in versions:
            if item['Name'] == self.version:
                return item['Id']

    def provision(self, params, name):
        servicecatalog = self._get_assumed_conduit_servicecatalog()
        is_provisioned = service_catalog.is_provisioned(name)
        if is_provisioned:
            print("Updating now...")
            service_catalog.update_provisioned(servicecatalog, self, name, params)
            print("Update success!")
        else:
            print("Provisioning now...")
            service_catalog.provision(servicecatalog, self, name, params)
            print("Provision success!")

    def terminate(self, name):
        servicecatalog = self._get_assumed_conduit_servicecatalog()
        is_provisioned = service_catalog.is_provisioned(name)
        if is_provisioned:
            print("Terminating now...")
            service_catalog.terminate_provisioned(servicecatalog, name)
        else:
            print("Artifact is not provisioned.")

    def _get_assumed_conduit_servicecatalog(self):
        sts = boto3.client('sts')
        account_id = sts.get_caller_identity().get('Account')
        print("Assumiing consuit IAM role...")
        creds = sts.assume_role(
            RoleArn='arn:aws:iam::{}:role/conduit/conduit-provisioner-role'.format(account_id),
            RoleSessionName='conduit-{}'.format(hash('conduit'))
        )
        servicecatalog = boto3.client(
            'servicecatalog',
            aws_access_key_id=creds['Credentials']['AccessKeyId'],
            aws_secret_access_key=creds['Credentials']['SecretAccessKey'],
            aws_session_token=creds['Credentials']['SessionToken'],
        )
        return servicecatalog

    def create_deployer_launch_constraint(self, portfolio, role_name):
        print("Creating Launch configuration...")
        response = service_catalog.list_product_constraints(portfolio.portfolio_id, self.product_id)
        exists = False
        for item in response:
            if item['Type'] == 'LAUNCH':
                print("Launch configuration exists, nothing to do.")
                exists = True
                break
        if not exists:
            role_arn = "arn:aws:iam::{}:role/conduit/{}".format(helper.get_account_id(), role_name)
            params = dict(
                RoleArn=role_arn
            )
            service_catalog.create_constraint(portfolio.portfolio_id, self.product_id, params, self.name)



import boto3

CFN = boto3.client('cloudformation')


def list_parameters(url):
    summary = CFN.get_template_summary(
        TemplateURL=url
    )
    if 'Parameters' in summary:
        return summary['Parameters']
    return []



import sys

import cmdln
from aws_conduit import conduit


class Conduit(cmdln.Cmdln):
    name = "conduit"

    def do_start(self, subcmd, opts):
        """
        ${cmd_name}: Set up the initial Conduit resources.

        ${cmd_usage}
        """
        conduit.configure()

    @cmdln.option("-e", "--email",
                  help="Contact email for product support.")
    @cmdln.option("-u", "--url",
                  help="Contact URL for product support.")
    @cmdln.option("-d", "--description",
                  help="Support information about the product.")
    def do_support(self, subcmd, opts):
        """
        ${cmd_name}: Set the default support configuration.

        ${cmd_usage}
        ${cmd_option_list}
        """
        conduit.set_default_support_config(
            description=opts.description,
            email=opts.email,
            url=opts.url
        )

    @cmdln.option("-n", "--name",
                  help="The name of the portfolio.")
    @cmdln.option("-d", "--description",
                  help="Information about the portfolio.")
    @cmdln.option("-i", "--id",
                  help="The portfolio id for use on an update.")
    def do_portfolio(self, subcmd, opts, action):
        """
        ${cmd_name}: Portfolio management for the masses!

        ${cmd_usage}
        Actions:
            create
            update
            delete
            list

        ${cmd_option_list}
        """
        actions = [
            'create',
            'update',
            'delete',
            'list'
        ]
        if action not in actions:
            raise ValueError("Not a valid action: {}".format(action))
        if action == 'create':
            conduit.new_portfolio(opts.name, opts.description)
        elif action == 'update':
            conduit.update_portfolio(opts.id, opts.name, opts.description)
        elif action == 'delete':
            conduit.delete_portfolio(opts.id)
        elif action == 'list':
            conduit.list_portfolios()
        else:
            print("{}: not a valid action for portfolio".format(action))

    @cmdln.option("-n", "--name",
                  help="The name of the portfolio.")
    @cmdln.option("-d", "--description",
                  help="Information about the product.")
    @cmdln.option("-c", "--cfntype",
                  help="yaml or json?")
    @cmdln.option("-p", "--portfolio",
                  help="The name of the portfolio.")
    @cmdln.option("-i", "--id",
                  help="The product id for use on an update.")
    @cmdln.option("-s", "--stackname",
                  help="The stack name when provisioning a prouct.")
    def do_product(self, subcmd, opts, action):
        """
        ${cmd_name}: Product management for the masses!

        ${cmd_usage}
        Actions:
            create
            update
            delete
            list
            associate
            provision

        ${cmd_option_list}
        """
        actions = [
            'create',
            'update',
            'delete',
            'list',
            'associate',
            'provision',
            'terminate'
        ]
        if action not in actions:
            raise ValueError("Not a valid action: {}".format(action))
        if action == 'create':
            conduit.new_product(opts.name, opts.description, opts.cfntype, opts.portfolio)
        elif action == 'update':
            conduit.update_product(opts.id, opts.name, opts.description, opts.cfntype)
        elif action == 'delete':
            conduit.delete_product(opts.id, None)
        elif action == 'list':
            conduit.list_products()
        elif action == 'associate':
            conduit.associate_product_with_portfolio(opts.id, opts.portfolio)
        elif action == 'provision':
            conduit.provision_product(opts.id, opts.name, opts.stackname)
        elif action == 'terminate':
            conduit.terminate_provisioned_product(opts.id, opts.stackname)
        else:
            print("{}: not a valid action for product".format(action))

    @cmdln.option("-p", "--product",
                  help="The name of the product to build.")
    def do_build(self, subcmd, opts, *action):
        """
        ${cmd_name}: Release a build from a conduitspec.yaml

        ${cmd_usage}
        Actions:
            major
            minor
            patch

        ${cmd_option_list}
        """
        if action and action[0] == 'major':
            print("Releasing new major version...")
            conduit.build('major', opts.product)
        elif action and action[0] == 'minor':
            print("Releasing new minor version...")
            conduit.build('minor', opts.product)
        elif action and action[0] == 'patch':
            print("Release new patch version...")
            conduit.build('patch', opts.product)
        else:
            print("Release new build version...")
            conduit.build('build', opts.product)

    @cmdln.option("-p", "--product",
                  help="The name of the product to provision.")
    @cmdln.option("-n", "--name",
                  help="A name to assign to the provisioned product.")
    def do_provision(self, subcmd, opts):
        """
        ${cmd_name}: Provision a product from a conduitspec.yaml

        ${cmd_usage}
        ${cmd_option_list}
        """
        conduit.provision_product_build(opts.product, opts.name)

    @cmdln.option("-n", "--name",
                  help="A name to assign to the provisioned product.")
    def do_terminate(self, subcmd, opts):
        """
        ${cmd_name}: Terminate a product from a conduitspec.yaml

        ${cmd_usage}
        ${cmd_option_list}
        """
        conduit.terminate_product(opts.name)

    def do_sync(seld, subcmd, opts):
        """
        ${cmd_name}: Ensure conduit is up to date.

        ${cmd_usage}
        ${cmd_option_list}
        """
        conduit.sync()

    @cmdln.option("-p", "--portfolio",
                  help="The name of the portfolio to package.")
    @cmdln.option("-e", "--environment",
                  help="The environment to package.")
    def do_package(self, subcmd, opts):
        conduit.package_portfolio(opts.portfolio, opts.environment)

    @cmdln.option("-p", "--portfolio",
                  help="The name of the portfolio to package.")
    @cmdln.option("-d", "--product",
                  help="The name of the product to package.")
    @cmdln.option("-e", "--environment",
                  help="The environment to package.")
    def do_package_product(self, subcmd, opts):
        conduit.package_product(opts.portfolio, opts.product, opts.environment)


def main():
    aws_conduit = Conduit()
    sys.exit(aws_conduit.main())
